#include <fstream.h>
#include <conio.h>
#include <time.h>
#include <math.h>

#define PERSPECT      0x00010000L

unsigned char buf[64][64];
unsigned char pal[1024];

struct Vector {
  long x, y, z;
  Vector() { }
  Vector(long x_, long y_, long z_)
  {
    x = x_;
    y = y_;
    z = z_;
  }
};

void Draw1stTriangle(long, long, long, long, long, Vector, Vector, Vector);
void DrawTriangle(Vector, Vector, Vector);

void Draw1stTriangle(long x1, long x2, long x3, long y1, long y2, Vector e1, Vector e2, Vector a)
{
  long a1, b1, c1, a2, b2, c2, d, u1, v1, u2, v2;
  long deltay = y2 - y1;
  long xx1 = x1 << 16, xx2 = x1 << 16, xx3, xx4, ii;
  long du, dv;
  unsigned long u, v;
  int i = y1, j, dy = 1;
  unsigned char c = 0x01;

  if (!deltay) return;

  if (deltay < 0)
  {
    dy = -1;
    deltay = -deltay;
  }

  long dx1 = ((x2 - x1) << 16) / deltay;
  long dx2 = ((x3 - x1) << 16) / deltay;

  for (; dy > 0 ? (i <= y2) : (i >= y2); i += dy)
  {
    for (j = xx1 >> 16; j <= (xx2 >> 16); j++)
    {
      a1 = (((long)j * e1.z) >> 16) - e1.x;
      b1 = (((long)j * e2.z) >> 16) - e2.x;
      c1 = a.x - ((a.z * (long)j) >> 16);
      a2 = (((long)i * e1.z) >> 16) - e1.y;
      b2 = (((long)i * e2.z) >> 16) - e2.y;
      c2 = a.y - ((a.z * (long)i) >> 16);
      d = a1 * b2 - a2 * b1;
      u1 = ((c1 * b2 - c2 * b1) << 6) / d;
      v1 = ((a1 * c2 - a2 * c1) << 6) / d;
      c = buf[u1 & 0x3F][v1 & 0x3F];
      asm  cmp    c,   7Fh
      asm  jna    l1
      asm  mov    c,   7Fh
l1:   asm  push   0A000h
      asm  pop    es
      asm  mov    bx,  i
      asm  mov    ax,  bx
      asm  shl    bx,  8
      asm  shl    ax,  6
      asm  add    bx,  ax
      asm  add    bx,  j
      asm  mov    di,  bx
      asm  mov    al,  c
      asm  stosb
    }
    xx1 += dx1;
    xx2 += dx2;
  }
}

void DrawTriangle(Vector v1, Vector v2, Vector v3)
{
  long xx1, xx3, xx4, dx2;
  long x1 = v1.x, x2 = v2.x, x3 = v3.x;
  long y1 = v1.y, y2 = v2.y, y3 = v3.y;
  Vector e1(x2 - x1, y2 - y1, (v2.z - v1.z) * PERSPECT);
  Vector e2(x3 - x1, y3 - y1, (v3.z - v1.z) * PERSPECT);
  Vector a(0, 0, v1.z * PERSPECT);
  long dy;

  if (y1 == y2 && y1 == y3) return;
  if (x1 == x2 && x1 == x3) return;

  if (y2 < y1)
  {
    y2 ^= y1; y1 ^= y2; y2 ^= y1;
    x2 ^= x1; x1 ^= x2; x2 ^= x1;
  }

  if (y3 < y1)
  {
    y3 ^= y1; y1 ^= y3; y3 ^= y1;
    x3 ^= x1; x1 ^= x3; x3 ^= x1;
  }

  if (y3 < y2)
  {
    y3 ^= y2; y2 ^= y3; y3 ^= y2;
    x3 ^= x2; x2 ^= x3; x3 ^= x2;
  }

  if (y2 == y3)
  {
    if (x2 > x3)
      Draw1stTriangle(x1, x3, x2, y1, y2, e1, e2, a);
    else
      Draw1stTriangle(x1, x2, x3, y1, y2, e1, e2, a);
  }
  else
  {
    dy = y3 - y1;
    if (dy < 0) dy = -dy;
    xx1 = x1;
    xx3 = x3;
    xx1 <<= 16;
    xx3 <<= 16;
    dx2 = (xx3 - xx1) / dy;
    dy = y2 - y1;
    if (dy < 0) dy = -dy;
    xx4 = (dx2 * dy) >> 16;
    xx4 += x1;
    if (xx4 < x2)
    {
      Draw1stTriangle(x1, xx4, x2, y1, y2, e1, e2, a);
      Draw1stTriangle(x3, xx4, x2, y3, y2, e1, e2, a);
    }
    else
    {
      Draw1stTriangle(x1, x2, xx4, y1, y2, e1, e2, a);
      Draw1stTriangle(x3, x2, xx4, y3, y2, e1, e2, a);
    }
  }
}

void SetGraphMode()
{
  asm  mov    ax,  13h
  asm  int    10h
}

void SetTextMode()
{
  asm  mov    ax,  03h
  asm  int    10h
}

void Vertret()
{
    asm  mov    dx,  03DAh
l1: asm  in     al,  dx
    asm  test   al,  8
    asm  jz     l1
l2: asm  in     al,  dx
    asm  test   al,  8
    asm  jnz    l2
}

void SetColor(unsigned char color, unsigned char r, unsigned char g, unsigned char b)
{
  asm  mov    dx,  3C8h
  asm  mov    al,  color
  asm  out    dx,  al
  asm  inc    dx
  asm  mov    al,  r
  asm  shr    al,  2
  asm  out    dx,  al
  asm  mov    al,  g
  asm  shr    al,  2
  asm  out    dx,  al
  asm  mov    al,  b
  asm  shr    al,  2
  asm  out    dx,  al
}

void SetPalette()
{
  for (int i = 1; i < 256; i++)
    SetColor(i, pal[(i << 2) + 2], pal[(i << 2) + 1], pal[i << 2]);
}

void LoadBMP(char *filename)
{
  ifstream in(filename, ios::binary);
  in.seekg(0x36, ios::beg);
  in.read(pal, sizeof(pal));
  in.read((char *)buf, sizeof(buf));
}

void Rotate(Vector s, Vector& d, int a)
{
  d.x = (s.x - 160) * cos(M_PI * 2 * a / 180) + (s.y - 100) * sin(M_PI * 2 * a / 180) + 160;
  d.y = (s.y - 100) * cos(M_PI * 2 * a / 180) - (s.x - 160) * sin(M_PI * 2 * a / 180) + 100;
  d.z = s.z;
}

void main()
{
  double end = 0;
  int a = 0;

  LoadBMP("texture.bmp");
  SetGraphMode();
  SetPalette();
  int i, count = 0;
  Vector v1(90, 30, 5), v2(230, 30, 5), v3(90, 170, 10), v4(230, 170, 3);
  Vector vv1, vv2, vv3, vv4;

  do
  {
    Rotate(v1, vv1, a);
    Rotate(v2, vv2, a);
    Rotate(v3, vv3, a);
    Rotate(v4, vv4, a);
    DrawTriangle(vv1, vv2, vv3);
//    DrawTriangle(vv4, vv3, vv2);
    count += 2;
//    a++;
    a %= 360;
  } while (!kbhit());

  end = clock() / CLK_TCK;
  SetTextMode();
  cout << "६ ࠡ ணࠬ: " << end << " .\n";
  cout << "⢮ 㣮쭨: " << count << '\n';
  cout << (double)(count / end) << " 㣮쭨 / ᥪ.\n";
  cprintf("Texture mapping - by HackMaster\n");
}
